/*
 ============================================================================
 BBgesture - A gesture recognition plugin for BB4win
 ============================================================================
 Created by MADant aka Kaloth - http://www.geocities.com/madant8
 ============================================================================

  Blackbox for Windows is free software and so is this plugin
  released under the GNU General Public License (GPL version 2 or later).
  
  You may mess with it as you wish! (But keep it open source please :))

 ============================================================================
*/

#ifndef _WIN32_WINNT
#define _WIN32_WINNT 0x0500
#endif

#ifdef __GNUC__
#define _WIN32_IE 0x0500
#define GetBlackboxPath _GetBlackboxPath
#define DLL_EXPORT //__declspec(dllexport)
#define LWA_ALPHA 2
#endif

#ifdef __BORLANDC__
#define DLL_EXPORT
#endif

#include "../../includes/BBApi.h"
#include <stdlib.h>
#include <math.h>

extern "C"
{
	DLL_EXPORT int beginPlugin(HINSTANCE hMainInstance);
	DLL_EXPORT void endPlugin(HINSTANCE hMainInstance);
	DLL_EXPORT LPCSTR pluginInfo(int field);
}

LPSTR szAppName         = "bbgesture";          // The name of our window class, etc.
LPSTR szVersion         = "bbgesture v0.5b";   // Used in MessageBox titlebars
LPSTR szInfoVersion     = "v0.5b";
LPSTR szInfoAuthor      = "Kaloth";
LPSTR szInfoRelDate     = "2004-03-23";
LPSTR szInfoLink        = "http://www.geocities.com/madant8";
LPSTR szInfoEmail       = "mulletwarriorextreem@hotmail.com";

//===========================================================================
HINSTANCE hInstance;
HWND hwndPlugin, hwndBlackbox, hwndGesture;
OSVERSIONINFO  osvinfo;

#define MAX_LINE_LENGTH 100

char rcpath[MAX_PATH];

#define PLUGIN_WINDOW_SIZE 20
int xpos, ypos;
int width = PLUGIN_WINDOW_SIZE, height = PLUGIN_WINDOW_SIZE;

#define GESTURE_WINDOW_SIZE 450
int g_xpos = 0, g_ypos = 0;
int g_width = GESTURE_WINDOW_SIZE, g_height = GESTURE_WINDOW_SIZE;

bool alwaysOnTop;
bool snapWindow;
bool hidden;
bool pluginToggle;
bool transparencyEnabled;
int transparencyAlpha;

bool usingWin2kXP;

char windowText[1000];

StyleItem myStyleItem;

int bevelWidth;
int borderWidth;
COLORREF borderColor;

void GetStyleSettings();
void ReadRCSettings();
void WriteRCSettings();

HFONT hFont;
HHOOK HKbHook;
HHOOK MHook;
HBITMAP g_bufbmp;
POINT pt;

int hotkey;
bool showPad;
bool showPadTips;
bool showToolSquare;
bool gesturing;
int oldx, oldy, d_oldx, d_oldy;
int b_oldx, b_oldy;
char gesture[MAX_LINE_LENGTH];

char gestures[25][MAX_LINE_LENGTH]; // 25 stored gestures.
char commands[25][MAX_LINE_LENGTH]; // 25 stored commands.
char arguments[25][MAX_LINE_LENGTH]; // 25 stored arguments.

void showGestureWindow(){
	if(showPad){
		ShowWindow(hwndGesture, SW_SHOW);
		InvalidateRect(hwndGesture, NULL, FALSE);
	}
}

void hideGestureWindow(){
	ShowWindow(hwndGesture, SW_HIDE);
}

void drawBackground(){
	PAINTSTRUCT ps; RECT r; HDC hdc, buf; HGDIOBJ otherbmp;
	//get window rectangle
	GetClientRect(hwndGesture, &r);
	
	//get the screen-buffer
	hdc = BeginPaint(hwndGesture, &ps);
	
	//first get a new 'device context'
	buf = CreateCompatibleDC(NULL);
	
	//select it into the DC and store, whatever there was before.
	otherbmp = SelectObject(buf, g_bufbmp);
	
	// Draw background gradient.
	MakeStyleGradient(buf, &r, &myStyleItem, true);
	
	DeleteObject(otherbmp);
	//delete the memory - 'device context'
	DeleteDC(buf);
	
	//done
	EndPaint(hwndGesture, &ps);
}

void deleteOffScreenBuffer(){
	//first get a new 'device context'
	HDC buf = CreateCompatibleDC(NULL);
			
	//select it into the DC and store, whatever there was before.
	HGDIOBJ otherbmp = SelectObject(buf, g_bufbmp);
			
	DeleteObject(SelectObject(buf, otherbmp));
			
	//delete the memory - 'device context'
	DeleteDC(buf);
}

void makeOffScreenBuffer(){
	PAINTSTRUCT ps; RECT r;
	// ----- Make the offscreen buffer ----- \\
	
	//get window rectangle
	GetClientRect(hwndGesture, &r);
	
	//get the screen-buffer
	HDC hdc = BeginPaint(hwndGesture, &ps);
	
	//then create a buffer in memory with the window size
	g_bufbmp = CreateCompatibleBitmap(hdc, r.right, r.bottom);
	
	//first get a new 'device context'
	HDC buf = CreateCompatibleDC(NULL);
	
	//select it into the DC and store, whatever there was before.
	HGDIOBJ otherbmp = SelectObject(buf, g_bufbmp);
	
	// Draw background gradient.
	MakeStyleGradient(buf, &r, &myStyleItem, true);
	
	DeleteObject(otherbmp);
	//delete the memory - 'device context'
	DeleteDC(buf);
	
	//done
	EndPaint(hwndGesture, &ps);		
}

void drawMouseMovement(int x1, int y1, int x2, int y2, COLORREF col, int width, char *str){
	PAINTSTRUCT ps; RECT r;
	
	//get window rectangle
	GetClientRect(hwndGesture, &r);
	
	//get the screen-buffer
	HDC hdc = BeginPaint(hwndGesture, &ps);
	
	//first get a new 'device context'
	HDC buf = CreateCompatibleDC(NULL);

	//select it into the DC and store, whatever there was before.
	HGDIOBJ otherbmp = SelectObject(buf, g_bufbmp);
	
	//draw transparently on our background-gradient
	SetBkMode(buf, TRANSPARENT);
	
	HPEN         hPen;
	HPEN         hPenOld;
	
	hPen = CreatePen(PS_SOLID, width, col);
	hPenOld = (HPEN)SelectObject(buf, hPen);
	
	MoveToEx(buf, x1 - g_xpos, y1 - g_ypos, NULL);
	LineTo(buf, x2 - g_xpos, y2 - g_ypos);
	
	if(str != NULL && showPadTips){
		r.left = x1 - g_xpos + 20;
		r.top = y1 - g_ypos;
		HGDIOBJ otherfont = SelectObject(buf, hFont);
		SetTextColor(buf, myStyleItem.TextColor);
		DrawText(buf, str, -1, &r, DT_TOP | DT_LEFT);
		SelectObject(buf, otherfont);
	}
	
	SelectObject(buf, hPenOld);
	DeleteObject(hPen);
	
	DeleteObject(otherbmp);
	//delete the memory - 'device context'
	DeleteDC(buf);
	
	//done
	EndPaint(hwndGesture, &ps);
	
	InvalidateRect(hwndGesture, NULL, FALSE);
}

int strpos(char *str1, char*str2){
	
}

// returns true if gest1 is within gest2.
float contains(char* gest1, char *gest2){
	if(strstr(gest1, gest2) != NULL){
		return 1.0; // its in there so 100% of it has been found.
	}
	
	return 0.0;
	
	/*// search for bits of gest1 in gest2...
	
	float scale = 0;
	char *ptr, *old = 0, *current;
	ptr = strtok(gest1, " "); // get turning points from gest1.
	while(ptr != NULL){
		if((current = strpos(gest2, ptr)) != NULL){ // find the turning point.
			if(old > current){ // is it further along the string than the previously found turning point?
				scale+=3; // we found a bit so add to the scale.
				old = current; // store a reference to this newly found turning point.
			}
			scale+=3;
		}
		ptr = strtok(NULL, " ");
	}
	if(scale>0){scale--;} // take off one for the removed whitespace.
	return (scale/(float)strlen(gest1)); // return the percentage of gest1 found in gest2.*/
}

// returns a percentage of how much gest1 is like gest2.
float rateGesture(char* gest1, char *gest2){
	//get the size ratio.
	float scale = (float)strlen(gest2)/(float)strlen(gest1);
	return scale * contains(gest1, gest2);
}

void runCommand(int n){
	char directory[MAX_LINE_LENGTH];
	char command[MAX_LINE_LENGTH];
	command[0] = '\0';
	strcpy(directory, commands[n]);
	
	if(strcmp(directory, "") != 0){
		int nLen;
		
		nLen = strlen(directory) - 1;
		while (nLen > 0 && directory[nLen] != '\\') nLen--;
		directory[nLen] = '\0';
		
		if(nLen == 0){
			strcpy(command, commands[n]);
		}else{
			int i;
			for(i = nLen + 1; i <= strlen(commands[n]); i++){
				command[i - nLen - 1] = commands[n][i];
			}
		}
	}
	
	char buffer[MAX_LINE_LENGTH];
	sprintf(buffer, "BBGesture Command -> %s", command);
	SendMessage(hwndBlackbox, BB_SETTOOLBARLABEL, 0, (LPARAM)buffer);
	
	BBExecute(NULL, "", command, arguments[n], directory, SW_SHOWNORMAL, true);
}

void processGesture(char *gest){
	if(strcmp(gest, "") != 0){
		int best = -1, n;
		float score, bestscore = 0;
		//score each gesture in comparison to this gets.
		for(n = 0; n < 25; n++){
			score = rateGesture(gest, gestures[n]);
			if(score >= 0.3){
				if(score > bestscore){
					bestscore = score;
					best = n;
				}
			}
		}
		if(best != -1){
			runCommand(best);
		}
	}
}

LRESULT CALLBACK MyMHook(int Code, WPARAM wParam, LPARAM lParam);

// The keyboard hook procedure.
LRESULT CALLBACK MyKBHook(int Code, WPARAM wParam, LPARAM lParam){
	if(Code >= 0){
		if(wParam == WM_KEYDOWN){
			KBDLLHOOKSTRUCT *key = (KBDLLHOOKSTRUCT *)lParam;
			if(key->vkCode == hotkey){
				if(!gesturing){
					MHook = SetWindowsHookEx(WH_MOUSE_LL, (HOOKPROC)MyMHook, GetModuleHandle(NULL), 0);
					deleteOffScreenBuffer();
					makeOffScreenBuffer();
					drawBackground();
					POINT coord;
					GetCursorPos(&coord);
					g_xpos = coord.x - (GESTURE_WINDOW_SIZE/2);
					g_ypos = coord.y - (GESTURE_WINDOW_SIZE/2);
					d_oldx = coord.x;
					d_oldy = coord.y;
					oldx = coord.x;
					oldy = coord.y;
					b_oldx = -1;
					b_oldy = -1;
					SetWindowPos(
						hwndGesture,
						HWND_TOPMOST,
						g_xpos, g_ypos, g_width, g_height, SWP_NOACTIVATE
						);
					gesture[0] = '\0';
					showGestureWindow();
					gesturing = true;
				}
			}
		}else if(wParam == WM_KEYUP){
			KBDLLHOOKSTRUCT *key = (KBDLLHOOKSTRUCT *)lParam;
			if(key->vkCode == VK_LWIN){
				if(gesturing){
					if (MHook != 0){UnhookWindowsHookEx(MHook);}
					MHook = 0;
					hideGestureWindow();
					gesturing = false;
					gesture[strlen(gesture)-1] = '\0'; // trim the extra space off the gesture string.
					char buffer[MAX_LINE_LENGTH];
					sprintf(buffer, "BBGesture -> |%s|", gesture);
					SendMessage(hwndBlackbox, BB_SETTOOLBARLABEL, 0, (LPARAM)buffer);
					processGesture(gesture);
				}
			}
		}
	}
	// Call the next hook in the chain.
	return CallNextHookEx(HKbHook, Code, wParam, lParam);
}

int distance(int x1, int y1, int x2, int y2){
	int d1 = (x1 - x2);
	int d2 = (y1 - y2);
	return sqrt(d1*d1 + d2*d2);
}

// The mouse hook procedure.
LRESULT CALLBACK MyMHook(int Code, WPARAM wParam, LPARAM lParam){
	if(Code >= 0){
		if(wParam == WM_MOUSEMOVE){
			MSLLHOOKSTRUCT *ms = (MSLLHOOKSTRUCT *)lParam;
			pt = (POINT)ms->pt;
			if(gesturing && distance(pt.x, pt.y, oldx, oldy) > 12){
				bool xdiff, ydiff, found = false;
				if(strlen(gesture) >= MAX_LINE_LENGTH)
					gesture[0] = '\0';
				
				// Detect the gesture!
				xdiff = (pt.x - oldx) > 0; // pos = true, neg = false;
				ydiff = (pt.y - oldy) > 0;
				if((pt.x - oldx) == 0){xdiff = -1;}
				if((pt.y - oldy) == 0){ydiff = -1;}
				
				if(ydiff != -1 || xdiff != -1){ // check if we are drawing a straight line or not.
				
					if(xdiff == b_oldx && xdiff != -1){
						if(xdiff && !(b_oldy == 1) && ydiff){// ^ right
							strcat(gesture, "tr ");
							drawMouseMovement(oldx, oldy, oldx, oldy, myStyleItem.ColorTo, 10, "tr");
							found = true;
						}else if(!xdiff && !(b_oldy == 1) && ydiff){// ^ left
							strcat(gesture, "tl ");
							drawMouseMovement(oldx, oldy, oldx, oldy, myStyleItem.ColorTo, 10, "tl");
							found = true;
						}else if(xdiff && (b_oldy == 1) && !ydiff){// v right
							strcat(gesture, "br ");
							drawMouseMovement(oldx, oldy, oldx, oldy, myStyleItem.ColorTo, 10, "br");
							found = true;
						}else if(!xdiff && (b_oldy == 1) && !ydiff){// v left
							strcat(gesture, "bl ");
							drawMouseMovement(oldx, oldy, oldx, oldy, myStyleItem.ColorTo, 10, "bl");
							found = true;
						}
					}
					
					if(ydiff == b_oldy && ydiff != -1){
						if(ydiff && (b_oldx == 1) && !xdiff){// > down
							strcat(gesture, "rd ");
							drawMouseMovement(oldx, oldy, oldx, oldy, myStyleItem.ColorTo, 10, "rd");
							found = true;
						}else if(!ydiff && (b_oldx == 1) && !xdiff){// > up
							strcat(gesture, "ru ");
							drawMouseMovement(oldx, oldy, oldx, oldy, myStyleItem.ColorTo, 10, "ru");
							found = true;
						}else if(ydiff && !(b_oldx == 1) && xdiff){// < down
							strcat(gesture, "ld ");
							drawMouseMovement(oldx, oldy, oldx, oldy, myStyleItem.ColorTo, 10, "ld");
							found = true;
						}else if(!ydiff && !(b_oldx == 1) && xdiff){// < up
							strcat(gesture, "lu ");
							drawMouseMovement(oldx, oldy, oldx, oldy, myStyleItem.ColorTo, 10, "lu");
							found = true;
						}
					}
					
				}
				
				drawMouseMovement(d_oldx, d_oldy, pt.x, pt.y, myStyleItem.TextColor, 3, NULL);
				d_oldx = pt.x;
				d_oldy = pt.y;
				oldx = pt.x;
				oldy = pt.y;
				b_oldx = xdiff;
				b_oldy = ydiff;
			}else{
				drawMouseMovement(d_oldx, d_oldy, pt.x, pt.y, myStyleItem.TextColor, 3, NULL);
				d_oldx = pt.x;
				d_oldy = pt.y;
			}
		}
	}
	// Call the next hook in the chain.
	return CallNextHookEx(MHook, Code, wParam, lParam);
}

// Function that is used to install the keyboard and mouse hooks.
void Hook(){
	HKbHook = SetWindowsHookEx(WH_KEYBOARD_LL, (HOOKPROC)MyKBHook, GetModuleHandle(NULL), 0);
	//MHook = SetWindowsHookEx(WH_MOUSE_LL, (HOOKPROC)MyMHook, GetModuleHandle(NULL), 0);
}

// Function that uninstalls all the hooks.
void Unhook(){
	if (HKbHook != 0){UnhookWindowsHookEx(HKbHook);}
	HKbHook = 0;
	
	if (MHook != 0){UnhookWindowsHookEx(MHook);}
	MHook = 0;
}

//===========================================================================

void set_window_modes(void)
{
	if (hidden && pluginToggle)
	{
		ShowWindow(hwndPlugin, SW_HIDE);
	}
	else
	{
		width = PLUGIN_WINDOW_SIZE, height = PLUGIN_WINDOW_SIZE;
		g_width = GESTURE_WINDOW_SIZE, g_height = GESTURE_WINDOW_SIZE;
		
		SetTransparency(
			hwndPlugin,
			transparencyEnabled ? (unsigned char)transparencyAlpha : 255
			);
			
		SetTransparency(
			hwndGesture,
			transparencyEnabled ? (unsigned char)transparencyAlpha : 255
			);

		SetWindowPos(
			hwndPlugin,
			alwaysOnTop ? HWND_TOPMOST : HWND_NOTOPMOST,
			xpos, ypos, width, height, SWP_NOACTIVATE|(showToolSquare ? SWP_SHOWWINDOW : NULL)
			);
			
		SetWindowPos(
			hwndGesture,
			HWND_TOPMOST,
			g_xpos, g_ypos, g_width, g_height, SWP_NOACTIVATE
			);

		InvalidateRect(hwndGesture, NULL, FALSE);
		InvalidateRect(hwndPlugin, NULL, FALSE);
	}
}

//===========================================================================
int beginPlugin(HINSTANCE hPluginInstance)
{
	int n;
	for(n = 0; n < 25; n++){
		strcpy(gestures[n], "");
		strcpy(commands[n], "");
		strcpy(arguments[n], "");
	}
	
	gesturing = false;
	LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
	LRESULT CALLBACK WndGestProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
	WNDCLASS wc;

	ZeroMemory(&osvinfo, sizeof(osvinfo));
	osvinfo.dwOSVersionInfoSize = sizeof (osvinfo);
	GetVersionEx(&osvinfo);
	usingWin2kXP =
		osvinfo.dwPlatformId == VER_PLATFORM_WIN32_NT
		&& osvinfo.dwMajorVersion > 4;

	hwndBlackbox = GetBBWnd();
	hInstance = hPluginInstance;

	// Register the window class...
	ZeroMemory(&wc,sizeof(wc));
	wc.lpfnWndProc = WndProc;           // our window procedure
	wc.hInstance = hPluginInstance;     // hInstance of .dll
	wc.lpszClassName = szAppName;       // our window class name
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.style = CS_DBLCLKS;

	if (!RegisterClass(&wc)) 
	{
		MessageBox(hwndBlackbox,
			"Error registering window class", szVersion,
				MB_OK | MB_ICONERROR | MB_TOPMOST);
		return 1;
	}

	// Get plugin and style settings...
	ReadRCSettings();
	GetStyleSettings();
	width = PLUGIN_WINDOW_SIZE, height = PLUGIN_WINDOW_SIZE;
	// Create the window...
	if (!(hwndPlugin = CreateWindowEx(
		usingWin2kXP ? WS_EX_TOOLWINDOW | WS_EX_LAYERED : WS_EX_TOOLWINDOW, // window style
		szAppName,                      // our window class name
		NULL,                           // NULL -> does not show up in task manager!                
		WS_POPUP
		| WS_CLIPCHILDREN
		| WS_CLIPSIBLINGS
		| (showToolSquare ? WS_VISIBLE : NULL),                   // window parameters
		xpos,                           // x position
		ypos,                           // y position
		width,                          // window width
		height,                         // window height
		NULL,                           // parent window
		NULL,                           // no menu
		hPluginInstance,                // hInstance of .dll
		NULL)))
	{
		MessageBox(0,
			"Error creating window", szVersion,
				MB_OK | MB_ICONERROR | MB_TOPMOST);
		return 1;
	}
	
	// Register the window class...
	ZeroMemory(&wc,sizeof(wc));
	wc.lpfnWndProc = WndGestProc;           // our window procedure
	wc.hInstance = hPluginInstance;     // hInstance of .dll
	wc.lpszClassName = "Gestbbgesture";       // our window class name
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.style = CS_DBLCLKS;

	if (!RegisterClass(&wc)) 
	{
		MessageBox(hwndBlackbox,
			"Error registering window class", szVersion,
				MB_OK | MB_ICONERROR | MB_TOPMOST);
		return 1;
	}
	
	g_width = GESTURE_WINDOW_SIZE, g_height = GESTURE_WINDOW_SIZE;
	g_xpos = 0, g_ypos = 0;
	// Create the gesture window...
	if (!(hwndGesture = CreateWindowEx(
		usingWin2kXP ? WS_EX_TOOLWINDOW | WS_EX_LAYERED : WS_EX_TOOLWINDOW, // window style
		"Gestbbgesture",                      // our window class name
		NULL,                           // NULL -> does not show up in task manager!                
		WS_POPUP
		| WS_CLIPCHILDREN
		| WS_CLIPSIBLINGS
		| WS_VISIBLE,                   // window parameters
		g_xpos,                           // x position
		g_ypos,                           // y position
		g_width,                          // window width
		g_height,                         // window height
		NULL,                           // parent window
		NULL,                           // no menu
		hPluginInstance,                // hInstance of .dll
		NULL)))
	{
		MessageBox(0,
			"Error creating gesture window", szVersion,
				MB_OK | MB_ICONERROR | MB_TOPMOST);
		return 1;
	}
	
	hideGestureWindow();
	
	// Hook the keyboard so we can observe the windows key.
	Hook();
	
	return 0;
}

//===========================================================================

void endPlugin(HINSTANCE hPluginInstance)
{
	// Remove the keyhook.
	Unhook();
	
	// Destroy the windows...
	DestroyWindow(hwndPlugin);
	DestroyWindow(hwndGesture);

	// Unregister window class...
	UnregisterClass(szAppName, hPluginInstance);
	
	// Unregister window class...
	UnregisterClass("Gestbbgesture", hPluginInstance);

	if (hFont)
		DeleteObject(hFont), hFont = NULL;
}

//===========================================================================

LPCSTR pluginInfo(int field)
{
	// pluginInfo is used by Blackbox for Windows to fetch information about
	// a particular plugin. At the moment this information is simply displayed
	// in an "About loaded plugins" MessageBox, but later on this will be
	// expanded into a more advanced plugin handling system. Stay tuned! :)
	switch (field)
	{
		case PLUGIN_NAME:
			return szAppName;       // Plugin name
		case PLUGIN_VERSION:
			return szInfoVersion;   // Plugin version
		case PLUGIN_AUTHOR:
			return szInfoAuthor;    // Author
		case PLUGIN_RELEASE:
			return szInfoRelDate;   // Release date, preferably in yyyy-mm-dd format
		case PLUGIN_LINK:
			return szInfoLink;      // Link to author's website
		case PLUGIN_EMAIL:
			return szInfoEmail;     // Author's email
		default:
			return szVersion;       // Fallback: Plugin name + version, e.g. "MyPlugin 1.0"
	}
}

//===========================================================================

void ShowMyMenu(bool popup)
{
	Menu *pMenu, *pSub;

	// We create the main plugin menu.
	// Each menu or submenu has an unique ID-string, which identifies it
	// for later updates.

	pMenu = MakeNamedMenu("bbgesture", "bbgesture_IDMain", popup);

	// --------------------------------------------------------------
	// ... and the configuration options menu, note the "internal" bro@ms!...
	pSub = MakeNamedMenu("Configuration", "bbgesture_IDConfig", popup);

	MakeMenuItem(pSub, "Always On Top", "@bbgestureInternal AlwaysOnTop", alwaysOnTop);
	MakeMenuItem(pSub, "Snap Window To Edge", "@bbgestureInternal SnapWindowToEdge", snapWindow);
	MakeMenuItem(pSub, "Toggle With Plugins", "@bbgestureInternal ToggleWithPlugins", pluginToggle);

	MakeMenuItem(pSub, "Transparency Enabled", "@bbgestureInternal TransparencyEnabled", transparencyEnabled);
	// insert an item to adjust an integer value. Implemented in bblean for now
	MakeMenuItemInt(pSub, "Alpha Value", "@bbgestureInternal TransparencyAlpha", transparencyAlpha, 0, 255);
	// ...and add it to the main plugin menu
	MakeSubmenu(pMenu, pSub, "Configuration");
	// Blank space...
	MakeMenuNOP(pMenu, NULL);
	// Edit the RC-File, using the default Blackbox Editor
	MakeMenuItem(pMenu, "Edit bbgesture.rc", "@bbgestureInternal Edit", false);
	// ...an "About" box is always nice, even though we now have the regular plugin info function...
	MakeMenuItem(pMenu, "About bbgesture", "@bbgestureInternal About", false);

	// And finally, we show the menu...
	ShowMenu(pMenu);
}

//===========================================================================

void toggle_setting(bool *flag, const char *toolbar_message, const char *rc_string)
{
	// toggle the setting
	*flag = false == *flag;

	// send a display sting to the toolbar
	char msg[80];
	sprintf(msg, "bbgesture -> %s %sabled!",
		toolbar_message, *flag ? "en" : "dis");
	SendMessage(hwndBlackbox, BB_SETTOOLBARLABEL, 0, (LPARAM)msg);

	// write the new setting to the rc - file
	WriteBool(rcpath, rc_string, *flag);

	// apply new settings
	set_window_modes();

	// and update the menu checkmarks
	ShowMyMenu(false);
}

//===========================================================================

LRESULT CALLBACK WndGestProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	char szTemp[MAX_PATH];

	// Blackbox messages we want to "subscribe" to:
	// BB_RECONFIGURE -> Sent when changing style and on reconfigure
	// BB_BROADCAST -> Broadcast message (bro@m)
	static int msgs[] = { BB_RECONFIGURE, BB_BROADCAST, 0};
	PAINTSTRUCT ps; RECT r; HDC hdc, buf; HGDIOBJ otherbmp;
	switch (message)
	{       
		// ==========
		case WM_CREATE:
			hwndGesture = hwnd;
		
			// Register to receive Blackbox messages...
			//SendMessage(hwndBlackbox, BB_REGISTERMESSAGE, (WPARAM)hwnd, (LPARAM)msgs);
			
			// make the window visible on all worspaces
			MakeSticky(hwnd);
			
			set_window_modes();
			
			makeOffScreenBuffer();
			
			break;

		// ==========
		case WM_DESTROY:
			deleteOffScreenBuffer();
		
			//remove us from the sticky-windows-list
			RemoveSticky(hwnd);

			// Unregister Blackbox messages...
			//SendMessage(hwndBlackbox, BB_UNREGISTERMESSAGE, (WPARAM)hwnd, (LPARAM)msgs);
			break;

		// ==========
		// Window redraw processing...
		case WM_PAINT:
		{

			//get the screen-buffer
			hdc = BeginPaint(hwnd, &ps);

			//get window rectangle
			GetClientRect(hwnd, &r);

			//to prevent flicker of the display, we draw to memory first,
			//then put it on screen in one single operation. This is like this:

			//first get a new 'device context'
			buf = CreateCompatibleDC(NULL);

			//select it into the DC and store, whatever there was before.
			otherbmp = SelectObject(buf, g_bufbmp);
			
			// use new api
			// Paint background according to the current style on the bitmap,
			// using the default border from style (true)
			//MakeStyleGradient(buf, &r, &myStyleItem, true);

			//select it into the DC and store, whatever there was before
			HGDIOBJ otherfont = SelectObject(buf, hFont);
			
			//draw transparently on our background-gradient
			SetBkMode(buf, TRANSPARENT);
			//... with some color
			SetTextColor(buf, myStyleItem.TextColor);
			//... at the right place
			r.top += (bevelWidth + borderWidth + 2);
			//...some text
			DrawText(buf, "BBGesture Pad", -1, &r, DT_CENTER);

			//restore the second previous whatever to the dc,

			SelectObject(buf, otherfont);

			//now copy the drawing from the paint buffer to the window...
			BitBlt(hdc,
				ps.rcPaint.left,
				ps.rcPaint.top,
				ps.rcPaint.right  - ps.rcPaint.left,
				ps.rcPaint.bottom - ps.rcPaint.top,
				buf,
				ps.rcPaint.left,
				ps.rcPaint.top,
				SRCCOPY
				);

			//restore the first previous whatever to the dc,
			//get in exchange back our bitmap, and delete it.
			DeleteObject(otherbmp);
			
			//delete the memory - 'device context'
			DeleteDC(buf);

			//done
			EndPaint(hwnd, &ps);
			break;
		}

		case WM_NCRBUTTONUP:
			ShowMyMenu(true);
			break;

		// ==========
		// The window is moving
		case WM_WINDOWPOSCHANGING:
			// Is SnapWindowToEdge enabled? - and the plugin not hidden?
			if (snapWindow && IsWindowVisible(hwnd))
			{
				// ...then snap window to screen edges, if the distance is
				// less than 10 pixel (if the last bool is false it uses the
				// current DesktopArea)
				SnapWindowToEdge((WINDOWPOS*)lParam, 10, true);
			}
			break;

		// ==========
		case WM_WINDOWPOSCHANGED:
		{
			break;
		}

		// ==========
		// Allow window to move...
		case WM_NCHITTEST:
			//return HTCAPTION;
			break;

		// ==========
		// Prevent closing with alt-f4
		case WM_CLOSE:
			break;

		// ==========
		// Let windows do everything else
		default:
			return DefWindowProc(hwnd,message,wParam,lParam);
	}
	// return zero, except from the hit-test message
	return 0;
}

//===========================================================================

LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	char szTemp[MAX_PATH];

	// Blackbox messages we want to "subscribe" to:
	// BB_RECONFIGURE -> Sent when changing style and on reconfigure
	// BB_BROADCAST -> Broadcast message (bro@m)
	static int msgs[] = { BB_RECONFIGURE, BB_BROADCAST, 0};
	
	switch (message)
	{       
		// ==========
		case WM_CREATE:
			hwndPlugin = hwnd;
		
			// Register to receive Blackbox messages...
			SendMessage(hwndBlackbox, BB_REGISTERMESSAGE, (WPARAM)hwnd, (LPARAM)msgs);
			
			// make the window visible on all worspaces
			MakeSticky(hwnd);
			
			set_window_modes();
			break;

		// ==========
		case WM_DESTROY:
			//remove us from the sticky-windows-list
			RemoveSticky(hwnd);

			// Unregister Blackbox messages...
			SendMessage(hwndBlackbox, BB_UNREGISTERMESSAGE, (WPARAM)hwnd, (LPARAM)msgs);
			break;

		// ==========
		// If Blackbox sends a reconfigure message, load the new style settings and
		// update the window...

		case BB_RECONFIGURE:
			ReadRCSettings();
			GetStyleSettings();
			set_window_modes();
			break;

		// ==========
		// Window redraw processing...
		case WM_PAINT:
		{
			PAINTSTRUCT ps; RECT r;

			//get the screen-buffer
			HDC hdc = BeginPaint(hwnd, &ps);

			//get window rectangle
			GetClientRect(hwnd, &r);

			//to prevent flicker of the display, we draw to memory first,
			//then put it on screen in one single operation. This is like this:

			//first get a new 'device context'
			HDC buf = CreateCompatibleDC(NULL);

			//then create a buffer in memory with the window size
			HBITMAP bufbmp = CreateCompatibleBitmap(hdc, r.right, r.bottom);

			//select it into the DC and store, whatever there was before.
			HGDIOBJ otherbmp = SelectObject(buf, bufbmp);

#if 1
			// use new api
			// Paint background according to the current style on the bitmap,
			// using the default border from style (true)
			MakeStyleGradient(buf, &r, &myStyleItem, true);
#else
			//Paint background according to the current style on the bitmap ...
			MakeGradient(buf, r,
				myStyleItem.type,
				myStyleItem.Color,
				myStyleItem.ColorTo,
				myStyleItem.interlaced,
				myStyleItem.bevelstyle,
				myStyleItem.bevelposition,
				bevelWidth,     // this is not used.
				borderColor,
				borderWidth
				);
#endif
			//select it into the DC and store, whatever there was before
			HGDIOBJ otherfont = SelectObject(buf, hFont);

			//draw transparently on our background-gradient
			SetBkMode(buf, TRANSPARENT);
			//... with some color
			SetTextColor(buf, myStyleItem.TextColor);
			//... at the right place
			r.top += (bevelWidth + borderWidth + 2);
			//...some text
			DrawText(buf, windowText, -1, &r, DT_CENTER);

			//restore the second previous whatever to the dc,

			SelectObject(buf, otherfont);

			//now copy the drawing from the paint buffer to the window...
			BitBlt(hdc,
				ps.rcPaint.left,
				ps.rcPaint.top,
				ps.rcPaint.right  - ps.rcPaint.left,
				ps.rcPaint.bottom - ps.rcPaint.top,
				buf,
				ps.rcPaint.left,
				ps.rcPaint.top,
				SRCCOPY
				);

			//restore the first previous whatever to the dc,
			//get in exchange back our bitmap, and delete it.
			DeleteObject(SelectObject(buf, otherbmp));

			//delete the memory - 'device context'
			DeleteDC(buf);

			//done
			EndPaint(hwnd, &ps);
			break;
		}

		// ==========
		// Broadcast messages (bro@m -> the bang killah! :D <vbg>)
		case BB_BROADCAST:
		{
			const char *msg_string = (LPCSTR)lParam;

			// First we check for the two "global" bro@ms,
			if (!stricmp(msg_string, "@BBShowPlugins"))
			{
				// @BBShowPlugins: Show window (global bro@m)
				hidden = false;
				set_window_modes();
				break;
			}
			if (!stricmp(msg_string, "@BBHidePlugins"))
			{
				// @BBHidePlugins: Hide window (global bro@m)
				hidden = true;
				set_window_modes();
				break;
			}

			// ==========
			// ...then we check for the plugin bro@ms that are available to
			// the end-user (ie. listed in the documentation! <g>)...
			if (!stricmp(msg_string, "@bbgestureBro@m"))
			{
				break;
			}

			// ==========
			// Finally, we check for our "internal" bro@m, which is used for
			// all plugin internal commands, for example the menu commands...

			static const char Example_Broam_ID[] = "@bbgestureInternal";

			if (!strnicmp(msg_string, Example_Broam_ID, sizeof Example_Broam_ID - 1))
			{
				msg_string += sizeof Example_Broam_ID - 1;
				while (*msg_string == ' ') msg_string ++;

				//...and check for the second part of the message-string
				if (!stricmp(msg_string, "AlwaysOnTop"))
				{
					// make use of our frienly ghost from above
					toggle_setting(&alwaysOnTop,        // the * to the bool to toggle
						msg_string,                     // the message to display on toolbar
						"bbgesture.plugin.alwaysontop:"   // the rc-setting to write
						);
					break;
				}
				if (!stricmp(msg_string, "SnapWindowToEdge"))
				{
					toggle_setting(&snapWindow,
						msg_string,
						"bbgesture.plugin.snapwindow:"
						);
					break;
				}
				if (!stricmp(msg_string, "TransparencyEnabled"))
				{
					toggle_setting(&transparencyEnabled,
						msg_string,
						"bbgesture.plugin.transparency.enabled:"
						);
					break;
				}

				{
					static char trans_alpha_id[] = "TransparencyAlpha ";
					if (!memicmp(msg_string, trans_alpha_id, sizeof(trans_alpha_id)-1))
					{
						transparencyAlpha = atoi(msg_string + sizeof(trans_alpha_id)-1);
						set_window_modes();
						WriteInt(rcpath, "bbgesture.plugin.transparency.alpha:", transparencyAlpha);
						break;
					}
				}

				if (!stricmp(msg_string, "ToggleWithPlugins"))
				{
					toggle_setting(&pluginToggle,
						msg_string,
						"bbgesture.plugin.togglewithplugins:"
						);
					break;
				}
				if (!stricmp(msg_string, "Edit"))
				{
					GetBlackboxEditor(szTemp);
					BBExecute(NULL, NULL, szTemp, rcpath, NULL, SW_SHOWNORMAL, false);
					break;
				}
				if (!stricmp(msg_string, "About"))
				{
					sprintf(szTemp,
						"%s"
						"\n\t"
						"\nhttp://www.geocities.com/madant8"
						"\nmulletwarriorextreem@hotmail.com",
						szVersion
						);
					MessageBox(
						NULL, szTemp, "About this plugin...",
						MB_OK | MB_ICONINFORMATION);
					break;
				}
			}
			break;
		}
		// ==========

		case WM_NCRBUTTONUP:
			ShowMyMenu(true);
			break;

		// ==========
		// The window is moving
		case WM_WINDOWPOSCHANGING:
			// Is SnapWindowToEdge enabled? - and the plugin not hidden?
			if (snapWindow && IsWindowVisible(hwnd))
			{
				// ...then snap window to screen edges, if the distance is
				// less than 10 pixel (if the last bool is false it uses the
				// current DesktopArea)
				SnapWindowToEdge((WINDOWPOS*)lParam, 10, true);
			}
			break;

		// ==========
		// Save window position if it changed...
		case WM_WINDOWPOSCHANGED:
		{
			WINDOWPOS* windowpos = (WINDOWPOS*)lParam;
			xpos = windowpos->x;
			ypos = windowpos->y;
			WriteInt(rcpath, "bbgesture.plugin.x:", xpos);
			WriteInt(rcpath, "bbgesture.plugin.y:", ypos);
			break;
		}

		// ==========
		// Allow window to move...
		case WM_NCHITTEST:
			return HTCAPTION;

		// ==========
		// Prevent closing with alt-f4
		case WM_CLOSE:
			break;

		// ==========
		// Let windows do everything else
		default:
			return DefWindowProc(hwnd,message,wParam,lParam);
	}
	// return zero, except from the hit-test message
	return 0;
}

//===========================================================================

void ReadRCSettings()
{
	char defaultpath[MAX_PATH]; int nLen, i;
	// First we look for the config file in the same folder as the plugin...
	GetModuleFileName(hInstance, rcpath, sizeof(rcpath));
	for (i=0;;i++)
	{
		nLen = strlen(rcpath); while (nLen && rcpath[nLen-1] != '\\') nLen--;
		// ...check the two possible filenames 'examplerc' and 'example.rc'...
		strcpy(rcpath+nLen, "bbgesturerc");  if (FileExists(rcpath)) break;
		strcpy(rcpath+nLen, "bbgesture.rc"); if (FileExists(rcpath)) break;
		if (0 == i)
		{   //...if not found aside the plugin, we keep this as the default path...
			strcpy(defaultpath, rcpath);
			//...and try the Blackbox directory.
			GetBlackboxPath(rcpath, sizeof(rcpath));
			continue;
		}
		//if again no config file was found, we use the default path...
		strcpy(rcpath, defaultpath);
		break; // ...and quit the loop
	}
	// If a config file was found we read the plugin settings from it...
	// ...if not, the ReadXXX functions give us just the defaults.
	xpos = ReadInt(rcpath, "bbgesture.plugin.x:", 10);
	ypos = ReadInt(rcpath, "bbgesture.plugin.y:", 10);
	if (xpos >= GetSystemMetrics(SM_CXSCREEN)) xpos = 10;
	if (ypos >= GetSystemMetrics(SM_CYSCREEN)) ypos = 10;
	transparencyEnabled 	= ReadBool(rcpath, "bbgesture.plugin.transparency.enabled:", false);
	transparencyAlpha   	= ReadInt(rcpath,  "bbgesture.plugin.transparency.alpha:", 60);
	alwaysOnTop        	= ReadBool(rcpath, "bbgesture.plugin.alwaysontop:", true);
	snapWindow          	= ReadBool(rcpath, "bbgesture.plugin.snapwindow:", true);
	pluginToggle        	= ReadBool(rcpath, "bbgesture.plugin.togglewithplugins:", true);
	hotkey 		    	= ReadInt(rcpath, "bbgesture.plugin.hotkey:", VK_LWIN);
	showPad			= ReadBool(rcpath, "bbgesture.plugin.pad.show:", true);
	showPadTips		= ReadBool(rcpath, "bbgesture.plugin.pad.tips:", true);
	showToolSquare		= ReadBool(rcpath, "bbgesture.plugin.toolSquare.show:", true);
	
	strcpy(windowText, "G");
	
	char buffer[30];
	int n = 0;
	for(n = 0; n < 25; n++){
		sprintf(buffer, "bbgesture.plugin.gesture%d:", n);
		strcpy(gestures[n], ReadString(rcpath,  buffer, ""));
		sprintf(buffer, "bbgesture.plugin.command%d:", n);
		strcpy(commands[n], ReadString(rcpath,  buffer, ""));
		sprintf(buffer, "bbgesture.plugin.argument%d:", n);
		strcpy(arguments[n], ReadString(rcpath,  buffer, ""));
	}
}

//===========================================================================
void GetStyleSettings()
{

	if (hFont)
		DeleteObject(hFont), hFont = NULL;

#if 1
	// -------------------------------------------------------------------
	// the direct way:
	// ---------------------

	myStyleItem = *(StyleItem*)GetSettingPtr(SN_TOOLBAR);

	bevelWidth  = *(int*)GetSettingPtr(SN_BEVELWIDTH);
	borderWidth = *(int*)GetSettingPtr(SN_BORDERWIDTH);
	borderColor = *(COLORREF*)GetSettingPtr(SN_BORDERCOLOR);

	hFont = CreateStyleFont(&myStyleItem);

#else
	// -------------------------------------------------------------------
	// the individual way:
	// -------------------------

	// Get the path to the current style file from Blackbox...
	const char *stylepath = stylePath();
	const char *temp;

	// Get the applicable color settings from the current style...
	myStyleItem.Color = ReadColor(stylepath, "toolbar.color:", "#000000");
	myStyleItem.ColorTo = ReadColor(stylepath, "toolbar.colorTo:", "#000000");
	myStyleItem.TextColor = ReadColor(stylepath, "toolbar.textColor:", "#FFFFFF");

	// ...gradient type, bevel etc. (using a StyleItem)...
	temp = ReadString(stylepath, "toolbar.label:", "Flat Gradient Vertical");
	ParseItem(temp, &myStyleItem);

	// ...font settings...
	strcpy(myStyleItem.Font, ReadString(stylepath, "toolbar.font:", "tahoma"));
	myStyleItem.FontHeight = ReadInt(stylepath, "toolbar.fontHeight:", 12);

	temp = ReadString(stylepath, "toolbar.fontHeight:", "normal");
	myStyleItem.FontWeight = stricmp(temp, "bold") ? FW_BOLD : FW_NORMAL;

	// ...and some additional parameters
	bevelWidth  = ReadInt(stylepath, "bevelWidth:",  1);
	borderWidth = ReadInt(stylepath, "borderWidth:", 1);
	borderColor = ReadColor(stylepath, "borderColor:", "#000000");

	//make the font
	hFont = CreateFont(
		myStyleItem.FontHeight, 0, 0, 0, myStyleItem.FontWeight,
		false, false, false,
		DEFAULT_CHARSET,
		OUT_DEFAULT_PRECIS,
		CLIP_DEFAULT_PRECIS,
		DEFAULT_QUALITY,
		DEFAULT_PITCH,
		myStyleItem.Font
		);
#endif

	// -------------------------------------------------------------------
	// calculate the window-size from the string, using the stylefont.

	RECT r = {0,0,0,0};
	HDC hdc = CreateCompatibleDC(NULL);
	HGDIOBJ otherfont = SelectObject(hdc, hFont);
	DrawText(hdc, windowText, -1, &r, DT_LEFT|DT_CALCRECT);

	int indent = (bevelWidth + borderWidth + 2);
	width  = r.right  + 2 * indent;
	height = r.bottom + 2 * indent;

	SelectObject(hdc, otherfont);
	DeleteDC(hdc);
}

//===========================================================================
